/*
 * Asterisk -- A telephony toolkit for Linux.
 * Copyright (C) 1999, Mark Spencer
 * 
 * Mark Spencer <markster@linux-support.net>
 *   
 * Say application 
 * 
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/config.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/say.h>
#include <asterisk/module.h>
#include <asterisk/manager.h>
#include <sys/types.h>
#include <regex.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>

static char *tdesc = "Say Application";

static char *app = "See comments in bug tracker.";

static char *synopsis = "...\n";

static char *descrip = "...\n";


STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

typedef struct {
  regex_t regex;
  char *pattern;
  char *symbol;
} rule;

typedef struct {
  char *language;
  int rules_count;
  rule *rules;
} language_rules;

static language_rules *lang_rules;

static int language_rules_count;

static char *replace(char *t, double *values) {

  char *buffer = malloc(2*strlen(t));
  if(!buffer) {
      ast_log(LOG_ERROR, "Out of memory\n");
      return NULL;
  }

  char *b = buffer;
  char *e;

  while((e=strstr(t, "/$"))!=NULL) {
    strncpy(b, t, e - t + 1);
    b += e - t + 1;
    b += sprintf(b, "%f", values[atoi(e+2)-1]);
    t = strchr(e, '}');
  }
  strcpy(b, t);  

  return buffer;  
}

static char *token(char **buffer) {

  static char *delimit(char **buffer, char dstart, char dend) {
    char *buf = *buffer;
    char *start = buf;

    while((*buf != dend) && (*buf != '\0') && (*buf != '\n')) 
      buf++;

    if(*buf=='\0' && dend!=' ') {          
      fprintf(stderr, "Can't find matching '%c' for '%c' at position: %d\n", dstart, dend, buf-*buffer);
      return NULL;
    }
    
    if(*buf!='\0') {
        *buf='\0';
        *buffer=buf+1;
    } else
        *buffer=buf;

    return start;
  }

  char *buf = *buffer;

  while(*buf != '\0') {
    if(*buf == '{') {
      *buffer=buf;
      return delimit(buffer, '{', '}');
    } else if(*buf != ' ') {
      *buffer=buf;
      return delimit(buffer, ' ', ' ');
    }
    buf++;
  }
  return NULL;
}

static int load_conf(language_rules **defs) {

  struct ast_config *cfg;  

  language_rules *lr = NULL;

  int lr_count = 0;

  cfg = ast_load("say.conf");
  if(!cfg) {
    ast_log(LOG_NOTICE, "No say config!");
    return -1;
  }

  char *cat = ast_category_browse(cfg, NULL);
  while(cat) {
    ast_log(LOG_DEBUG, "Rules for language '%s':\n", cat);
    lr = (language_rules*)realloc(lr, sizeof(language_rules) * (lr_count + 1));
    if(!lr) {
	  ast_log(LOG_ERROR, "Out of memory\n");
	  return -1;
    }

    language_rules *cr = lr + lr_count;
    cr->language = (char*)malloc(strlen(cat)+1);

    if(!cr->language) {
	  ast_log(LOG_ERROR, "Out of memory\n");
	  return -1;
    }

    strcpy(cr->language, cat);

    cr->rules = NULL;
    cr->rules_count = 0;
    struct ast_variable *r = ast_variable_browse(cfg, cat);
    while(r) {
      cr->rules = realloc(cr->rules, sizeof(rule) * (cr->rules_count + 1));           
      if(!cr->rules) {
	ast_log(LOG_ERROR, "Out of memory\n");
	return -1;
      }
      if(regcomp(&(cr->rules[cr->rules_count].regex), r->name, REG_EXTENDED|REG_NOSUB) != 0) {
        ast_log(LOG_NOTICE, "Can't parse re(%s).", r->name);
        /*should free memory here*/
        return 0; 
      }
      cr->rules[cr->rules_count].pattern = (char*)malloc(strlen(r->name)+1);
      if(!cr->rules[cr->rules_count].pattern) {
	ast_log(LOG_ERROR, "Out of memory\n");
	return -1;
      }
      strcpy(cr->rules[cr->rules_count].pattern, r->name);
      cr->rules[cr->rules_count].symbol = (char*)malloc(strlen(r->value)+1);
      if(!cr->rules[cr->rules_count].symbol) {
	ast_log(LOG_ERROR, "Out of memory\n");
	return -1;
      }
      strcpy(cr->rules[cr->rules_count].symbol, r->value);
      ast_log(LOG_DEBUG, "Rule(%d): %s -> %s.\n", cr->rules_count + 1,  cr->rules[cr->rules_count].pattern, cr->rules[cr->rules_count].symbol);
      cr->rules_count++;
      r = r->next;
    }

    lr_count++;
    cat = ast_category_browse(cfg, cat);
  }
  *defs = lr;
  return lr_count;
}


static int say_exec(struct ast_channel *chan, void *data)
{

  static int play_file(char *fn) {    
    ast_log(LOG_DEBUG, "Trying to play: %s\n", fn);
    int res = ast_streamfile(chan, fn, chan->language);
    if(!res) {
	res = ast_waitstream(chan, "");
    }
    ast_stopstream(chan);
    return res;
  }

  int res = 0;

  struct localuser *u;

  if (!data) {
    ast_log(LOG_WARNING, "Say requires an argument (phrase to say|numbers to include in phrase)\n");
    return -1;
  }

  /* parsing parameters... */

  int BUFFER_SIZE = 1000;     
  char buf[BUFFER_SIZE];

  strncpy(buf, (char *)data, BUFFER_SIZE);
  buf[BUFFER_SIZE-1] = '\0';
 
  char *sentence = buf;
  char *numbers;

  int values_count = 0;
  double *values = NULL;   

  if(sentence) {
    numbers = strchr(sentence, '|');
    if(numbers) {
      *numbers = '\0';
      numbers++;
      char *last;
      do {
        errno = 0;
	double value = strtod(numbers, &last);
        if(errno == 0) {
	  values = realloc(values, (values_count+1)*sizeof(double));
	  if(!values) {
	    ast_log(LOG_ERROR, "Out of memory\n");
	    return -1;
	  }
	  values[values_count] = value;
          values_count++;
	} else {
          ast_log(LOG_WARNING, "Can't parse value list at %s\n", last);
	  return -1;
	}
        if(*last != '|') {
	  last = strchr(last, '|');
	}
        numbers = last+1;
      } while(last!='\0' && last != NULL);
    }
  }

  LOCAL_USER_ADD(u);

  /* Looking for rule set for current language... */

  language_rules *lr = NULL;
  int i;
  for(i=0; i<language_rules_count; i++) {
    if(strcmp(lang_rules[i].language, chan->language)==0)
      lr = &lang_rules[i];
  }

  if(lr==NULL) {
      ast_log(LOG_NOTICE, "Can't find rules for language: %s. Defaults to: %s\n", chan->language, lang_rules[0].language);
      lr=&lang_rules[0];
  }

  /* substituting $?  with values... */
  char *tmp =  replace(sentence, values);

  /* applying rules and playing resulting files... */
  char *t;
  while((t=token(&tmp))!= NULL) {
    ast_log(LOG_DEBUG, "t: %s\n", t);
    if(*t=='{') {
      int i;
      for(i = 0; i < lr->rules_count; i++) {
	if(regexec(&lr->rules[i].regex, t+1, (size_t) 0, NULL, 0)==0) {
	  if(strcmp(lr->rules[i].symbol, "number(?)")==0) {
	    int x = atoi(strstr(t+1, "/")+1); /* saying doubles should be possible  */
            ast_log(LOG_DEBUG, "x: %d\n", x);
            res = ast_say_number(chan, x, "", chan->language, "");            
	  /*
	  } if(strcmp(lr->rules[i].symbol, "currency(?)")==0) {
              TODO
          */
          } else {
            res = play_file(lr->rules[i].symbol);  
	  }
          break;
	}
      }
    } else {
      res = play_file(t);  
    }
    if(res==-1) return res;
  }

  LOCAL_USER_REMOVE(u);
  return res;
}


static int reload_conf(void)
{
  if(lang_rules!=NULL) {
    int i;
    for(i = 0; i < language_rules_count; i++) {
      free(lang_rules[i].language);
      free(lang_rules[i].rules);
    }
    free(lang_rules);
  }
      
  language_rules_count =  load_conf(&lang_rules);
  return language_rules_count < 0;
}


int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}


int reload(void)
{	
	return reload_conf();
}

int load_module(void)
{
	int res;
        res = ast_register_application(app, say_exec, synopsis, descrip);
        if(!res)
            res = reload_conf();
	return res;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

